#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <unistd.h>
#include <limits.h>
#include <stddef.h>

static char const prog[]="caldelay";

static char const var_weekday[]="$WEEKDAY";
static char const var_year[]="$YEAR";
static char const var_month[]="$MONTH";
static char const var_monthday[]="$MONTHDAY";
static char const var_hour[]="$HOUR";
static char const var_minute[]="$MINUTE";
static char const var_second[]="$SECOND";

static int increased=0;
static int need_increase=0;
struct caltime ct;

static unsigned long limit_weekday(void) { return  7; }
static unsigned long limit_year   (void) { return -1; }
static unsigned long limit_month  (void) { return 12; }
static unsigned long limit_monthday(void) {
  static unsigned long const limits[]={
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
  };
  unsigned long leap;
  if (ct.date.month!=1)
    return limits[ct.date.month];
  leap=ct.date.year;
  if (leap%4!=0) leap=0;
  else {
    leap/=4;
    if (leap%25!=0) leap=1;
    else {
      leap/=25;
      if (leap%4!=0) leap=0;
      else leap=1;
    }
  }
  return 28+leap;
}
static unsigned long limit_hour  (void) { return 24; }
static unsigned long limit_minute(void) { return 60; }
static unsigned long limit_second(void) { return 60; }

typedef enum { eq, div } spectype;
typedef struct {
  char const* const env;
  spectype type;
  unsigned long spec;
  unsigned long value;
  unsigned long (* const limit)(void);
} unit;
static unit weekday_units[]={
  { var_weekday,  -1, -1, -1, limit_weekday  },
  { var_hour,     -1, -1, -1, limit_hour     },
  { var_minute,   -1, -1, -1, limit_minute   },
  { var_second,   -1, -1, -1, limit_second   }
};
static unit monthday_units[]={
  { var_year,     -1, -1, -1, limit_year     },
  { var_month,    -1, -1, -1, limit_month    },
  { var_monthday, -1, -1, -1, limit_monthday },
  { var_hour,     -1, -1, -1, limit_hour     },
  { var_minute,   -1, -1, -1, limit_minute   },
  { var_second,   -1, -1, -1, limit_second   }
};

static void do_unit(unit* const u) {
  unsigned long value=(increased? 0: u->value);
  unsigned long const spec=u->spec;
  spectype const type=u->type;
  if (type==eq) {
    if (need_increase) return;
    if (value==spec) goto setval;
    if (value>spec) goto need_increase;
    value=spec;
  } else /* type==div */ {
    if (need_increase) {
      value+=spec;
      if (value>=u->limit()) return;
      increased=1;
      need_increase=0;
      goto setval;
    }
    if (value==0) goto setval;
    if (spec==0) goto need_increase;
    if (value%spec==0) goto setval;
    value+=spec;
    value-=value%spec;
    if (value>=u->limit()) goto need_increase;
  }
  increased=1;
  setval:
  u->value=value;
  return;
  need_increase:
  need_increase=1;
  return;
}

int main(int argc, char** argv) {
  struct tai now;
  unsigned long delay;
  int weekday;
  char const* x;
  unit* units;
  size_t n_units;
  size_t i;
  char delaybuf[(sizeof delay*CHAR_BIT+2)/3+1];

  if (argc<2)
    die4(x, 111, prog, usage, prog, " program [arg ...]");

  x=envget(var_weekday);
  if (x) {
    units=weekday_units;
    n_units=sizeof weekday_units/sizeof (unit);
  } else {
    units=monthday_units;
    n_units=sizeof monthday_units/sizeof (unit);
  }

  {
    int seen_any=0;
    for (i=0; i!=n_units; ++i) {
      x=envget(units[i].env);
      if (x!=null) {
        unsigned int len;
        seen_any=1;
        switch (*x) {
          case '=': units[i].type=eq;  break;
          case '/': units[i].type=div; break;
          default:
            die4(x, 111, prog, fatal, units[i].env, err_malformed);
        }
        len=scan_ulong(x+1, &units[i].spec);
        if (len==0 || x[1+len]!='\0')
          die4(x, 111, prog, fatal, units[i].env, err_malformed);
      } else if (seen_any) {
        units[i].type=eq;
        units[i].spec=0;
      } else {
        units[i].type=div;
        units[i].spec=1;
      }
    }
  }

  tai_now(&now);
  caltime_utc(&ct, &now, &weekday, null);

  if (units==weekday_units) {
    units[0].value=weekday;
    units[1].value=ct.hour;
    units[2].value=ct.minute;
    units[3].value=ct.second;
  } else {
    units[0].value=ct.date.year;
    units[1].value=ct.date.month-1;
    units[2].value=ct.date.day-1;
    units[3].value=ct.hour;
    units[4].value=ct.minute;
    units[5].value=ct.second;
  }

  for (i=0; i!=n_units;) {
    do_unit(&units[i]);
    while (need_increase) {
      if (i==0) {
        if (units==monthday_units)
          die3(x, 111, prog, fatal, "all run times are in the past");
        units[0].value=0;
        ct.date.day+=7;
        continue;
      }
      --i;
      do_unit(&units[i]);
    }
    ++i;
  }

  if (units==weekday_units) {
    ct.date.day +=units[0].value;
    ct.date.day -=weekday;
    ct.hour      =units[1].value;
    ct.minute    =units[2].value;
    ct.second    =units[3].value;
  } else {
    ct.date.year =units[0].value;
    ct.date.month=units[1].value+1;
    ct.date.day  =units[2].value+1;
    ct.hour      =units[3].value;
    ct.minute    =units[4].value;
    ct.second    =units[5].value;
  }

  {
    unsigned char buf[TAI_PACK];
    struct tai then;
    caltime_tai(&ct, &then);
    tai_sub(&then, &then, &now);
    tai_pack(buf, &then);
    delay=0;
    for (i=0; i<TAI_PACK; ++i) {
      unsigned long const d=delay<<CHAR_BIT;
      if (d>>CHAR_BIT!=delay)
        die3(x, 111, prog, fatal, err_overflow);
      delay=d+buf[i];
      if (delay<d)
        die3(x, 111, prog, fatal, err_overflow);
    }
  }

  delaybuf[fmt_ulong(delaybuf, delay)]='\0';
  pathexec_env((char*)var_delay+1, delaybuf);
  ++argv;
  pathexec(argv);
  die5(sys, 111, prog, fatal, err_exec, argv[0], syserr);
}
