#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <signal.h>

static char const prog[]="delayrun";

static void alarm_handler(int sig) {
  (void)sig;
}

int main(int argc, char** argv) {
  unsigned long delay=0;
  char const* x;

  x=envget(var_delay);
  if (x!=null) {
    unsigned int const len=scan_ulong(x, &delay);
    if (len==0 || x[len]!='\0')
      die4(x, 111, prog, fatal, var_delay, err_malformed);
    if (delay>LONG_MAX)
      die3(x, 111, prog, fatal, err_overflow);
  }

  if (x==null || delay!=0) {
    struct itimerval timer;
    sigset_t oldset, alarmset, suspendset;
    struct sigaction oldact, alarmact;
    timer.it_interval.tv_sec=0;
    timer.it_interval.tv_usec=0;
    timer.it_value.tv_sec=delay;
    timer.it_value.tv_usec=0;
    if (sigemptyset(&alarmset)!=0)
      die3(sys, 111, prog, fatal, err_sigempty);
    if (sigaddset(&alarmset, SIGALRM)!=0)
      die3(sys, 111, prog, fatal, err_sigadd);
    if (sigprocmask(SIG_BLOCK, null, &suspendset)!=0)
      die3(sys, 111, prog, fatal, err_sigmask);
    if (sigprocmask(SIG_BLOCK, &alarmset, &oldset)!=0)
      die3(sys, 111, prog, fatal, err_sigmask);
    if (setitimer(ITIMER_REAL, &timer, null)!=0)
      die3(sys, 111, prog, fatal, err_timer);
    alarmact.sa_handler=alarm_handler;
    if (sigemptyset(&alarmact.sa_mask)!=0)
      die3(sys, 111, prog, fatal, err_sigempty);
    alarmact.sa_flags=0;
    if (sigaction(SIGALRM, &alarmact, &oldact)!=0)
      die3(sys, 111, prog, fatal, err_sigact);
    if (sigdelset(&suspendset, SIGALRM)!=0)
      die3(sys, 111, prog, fatal, err_sigdel);
    if (sigsuspend(&suspendset)!=0 && errno!=EINTR)
      die3(sys, 111, prog, fatal, err_suspend);
    if (sigaction(SIGALRM, &oldact, null)!=0)
      die3(sys, 111, prog, fatal, err_sigact);
    if (sigprocmask(SIG_SETMASK, &oldset, null)!=0)
      die3(sys, 111, prog, fatal, err_sigmask);
  }

  if (argc<2) _exit(0);
  ++argv;
  pathexec_run(argv[0], argv, environ);
  die5(sys, 111, prog, fatal, err_exec, argv[0], syserr);
}
