#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <limits.h>

static char const prog[]="maxdelay";

static char const var_lastrun[]="$LASTRUN";
static char const var_maxinterval[]="$MAXINTERVAL";

int main(int argc, char** argv) {
  struct stat statbuf;
  unsigned long delay;
  unsigned long max;
  char const* lastrun;
  char const* x;
  time_t const now=time(null);
  unsigned int len;
  char delaybuf[(sizeof delay*CHAR_BIT+2)/3+1];

  if (argc<2)
    die4(x, 111, prog, usage, prog, " program [arg ...]");

  lastrun=envget(var_lastrun);
  if (lastrun==null)
    die4(x, 111, prog, fatal, var_lastrun, err_notset);

  x=envget(var_delay);
  if (x==null)
    die4(x, 111, prog, fatal, var_delay, err_notset);
  len=scan_ulong(x, &delay);
  if (len==0 || x[len]!='\0')
    die4(x, 111, prog, fatal, var_delay, err_malformed);

  x=envget(var_maxinterval);
  if (x==null)
    die4(x, 111, prog, fatal, var_maxinterval, err_notset);
  len=scan_ulong(x, &max);
  if (len==0 || x[len]!='\0')
    die4(x, 111, prog, fatal, var_maxinterval, err_malformed);

  if (stat(lastrun, &statbuf)!=0) {
    if (errno!=ENOENT)
      die5(sys, 111, prog, fatal, err_stat, lastrun, syserr);
  } else if (statbuf.st_mtime<=now) {
    unsigned long since=now-statbuf.st_mtime;
    if (since>=max) delay=0;
    else {
      since+=delay;
      if (since<delay || since>max) delay+=max-since;
    }
  } else {
    max+=(unsigned long)(statbuf.st_mtime-now);
    if (delay>max) delay=max;
  }

  delaybuf[fmt_ulong(delaybuf, delay)]='\0';
  pathexec_env((char*)var_delay+1, delaybuf);
  ++argv;
  pathexec(argv);
  die5(sys, 111, prog, fatal, err_exec, argv[0], syserr);
}
